"""This module contains Equatorial class.
"""
import math

from astro_toolbox.angle.hms import AngleHMS
from astro_toolbox.angle.dms import AngleDMS
from astro_toolbox.angle.degrees import AngleDeg
from astro_toolbox.angle.radians import AngleRad
from astro_toolbox.coordinates.location import Location
from astro_toolbox.time.core import AstroDateTime

class Equatorial():
    """This class represents astronomical equatorial coordinate system.

    Attributes
    ----------
    alpha : tuple
        Object right-ascension.
    delta : tuple
        Object declination.
    name : str
        Object name.
    magnitude : float
        Object magnitude.
    """
    def __init__(self, alpha: tuple | str,
                delta: tuple | str,
                name: str = None,
                magnitude: float = None):
        """Constructor method

        Parameters
        ----------
        alpha : tuple | str
            Object right-ascension as tuple or str (``dd:dd:dd.dd`` or ``ddhddmdd.dds``).
        delta : tuple | str
            Object declination as tuple or str (``dd:dd:dd.dd`` or ``dd°dd'dd.dd"``).
        name : str, optional
            Object name, by default None.
        magnitude : float
            Object magnitude, by default is None.
        """
        self.name = name
        self.alpha = AngleHMS(alpha)
        self.delta = AngleDMS(delta)
        self.magnitude = magnitude

    def __repr__(self):
        """Representative method.

        Returns
        -------
        str
            Return a class representative string.
        """
        return (f'{self.name}: \N{GREEK SMALL LETTER ALPHA} = {self.alpha} ' +
                f'\N{GREEK SMALL LETTER DELTA} = {self.delta} '
                f'v = {self.magnitude}')

    def get_hourangle(self, gamma: tuple | str):
        """Right-Ascension to Hour-Angle converting method.

        .. math:: HA = RA - \\gamma

        Parameters
        ----------
        gamma : tuple | str
            Sidereal Time angle as tuple or string.

        Returns
        -------
        tuple
            Hour-Angle tuple as tuple.
        """
        gamma_angle = AngleHMS(gamma)
        hour_angle = AngleDeg(gamma_angle.hmstodeg() - self.alpha.hmstodeg())
        return hour_angle.degtohms()

    def calculate_airmass(self, gamma: tuple | str, location: Location):
        """Airmass calculation method.
        The airmass is calculate with the Pickering (2002) formula from DIO,
        The International Journal of Scientific History vol. 12.

        .. math:: X = \\frac{1}{sin(h+\\frac{244}{165+47h^{1.1}})}

        For altitude angle calculation c.f. to_horizontal method.

        Parameters
        ----------
        gamma : tuple | str
            Sidereal Time angle in hms.
        location : Location
            observer location.

        Returns
        -------
        float
            Airmass value of the object.
        """
        lat = location.latitude.dmstorad()
        hour_angle = AngleHMS(self.get_hourangle(gamma=gamma)).hmstorad()
        altitude = AngleRad(math.asin(math.cos(lat) * math.cos(hour_angle) *
                    math.cos(self.delta.dmstorad()) + math.sin(lat) *
                    math.sin(self.delta.dmstorad()))).radtodeg()
        if altitude < 0:
            return 40
        return 1/(math.sin(AngleDeg(altitude + 244/(165 + 47 * altitude ** 1.1)).degtorad()))

    def to_horizontal(self, gamma: tuple | str, location: Location):
        """Equatorial to Horizontal converting method.

        .. math:: h=sin^{-1}(cos\\Phi cos H cos\\delta+sin\\Phi sin\\delta)

        .. math:: A=sin^{-1}(\\frac{-sin H cos \\delta}{cosh})

        Parameters
        ----------
        gamma : tuple | str
             Sidereal Time angle as tuple or string.
        location : Location
            observer location.

        Returns
        -------
        tuple
            Tuple containing two tuple in DMS with (azimuth, altitude).
        """
        lat = location.latitude.dmstorad()
        hour_angle = AngleHMS(self.get_hourangle(gamma=gamma)).hmstorad()
        altitude = AngleRad((math.asin(math.cos(lat) * math.cos(hour_angle) *
                            math.cos(self.delta.dmstorad())
            + math.sin(lat) * math.sin(self.delta.dmstorad()))))
        azimuth = AngleRad((math.asin(-math.sin(hour_angle) *
            math.cos(self.delta.dmstorad()) / math.cos(altitude.anglevalue)))%(2*math.pi))
        return (azimuth.radtodms(), altitude.radtodms())

    def compute_on_date_coord(self, year: float):
        """On date Equatorial coordinates calculation method from J2000 Equatorial coordinate.

        .. math:: \\Delta year = \\frac{(year-2000)}{100}

        .. math:: M=1.2812323\\Delta year+0.0003879\\Delta year^2+0.0000101\\Delta year^3

        .. math:: N=0.5567530\\Delta year-0.0001185\\Delta year^2+0.0000116\\Delta year^3

        .. math:: \\Delta \\alpha=M+Nsin\\alpha tan\\delta

        .. math:: \\Delta \\delta=Ncos\\alpha

        .. math:: \\alpha=\\alpha_{J2000}+\\Delta \\alpha

        .. math:: \\delta=\\delta_{J2000}+\\Delta \\delta

        Parameters
        ----------
        year : float
            Current year (months and days can be counted as year fraction).
        """
        var_year = (year - 2000.0)/100
        m_coeff = (1.2812323 * var_year + 0.0003879 *
                var_year**2 + 0.0000101 * var_year**3) * math.pi/180
        n_coeff = (0.5567530 * var_year - 0.0001185 *
                var_year**2 + 0.0000116 * var_year**3) * math.pi/180
        var_alpha = (m_coeff + n_coeff * math.sin(self.alpha.hmstorad()) *
                    math.tan(self.delta.dmstorad()))
        var_delta = n_coeff * math.cos(self.alpha.hmstorad())
        alpha = AngleHMS(AngleRad(self.alpha.hmstorad() + var_alpha).radtohms())
        delta = AngleDMS(AngleRad(self.delta.dmstorad() + var_delta).radtodms())
        self.alpha = alpha
        self.delta = delta

    def calculate_rise_time(self, location:Location, date: tuple | str, altitude_0: float=0.0):
        """Rise time calculation method.

        Parameters
        ----------
        location : Location
            Observer location as Location class.
        date : tuple | str
            Date as tuple or str (``yyyy-mm-dd`` or ``yyyy:mm:dd``).

        Returns
        -------
        tuple
            Tuple containing the rising time in HMS.
        """
        time = AstroDateTime(date)
        hour_angle = math.acos((math.sin(altitude_0*math.pi/180) -
                    math.sin(location.latitude.dmstorad()) *
                    math.sin(self.delta.dmstorad())) /
                    (math.cos(location.latitude.dmstorad()) *
                    math.cos(self.delta.dmstorad())))*180/math.pi
        tsg0 = AngleHMS(AstroDateTime(time.date+(0 , 0, 0)).get_gmst()).hmstodeg()
        return AngleDeg((self.alpha.hmstodeg() -
                                hour_angle +
                                location.longitude.dmstodeg() -
                                tsg0) /
                                1.002737909).degtohms()

    def calculate_set_time(self, location:Location, date: tuple | str, altitude_0: float=0.0):
        """Set time calculation method.

        Parameters
        ----------
        location : Location
            Observer location as Location class.
        date : tuple | str
            Date as tuple or str (``yyyy-mm-dd`` or ``yyyy:mm:dd``).

        Returns
        -------
        tuple
            Tuple containing the setting time in HMS.
        """
        time = AstroDateTime(date)
        hour_angle = math.acos((math.sin(altitude_0*math.pi/180) -
                    math.sin(location.latitude.dmstorad()) *
                    math.sin(self.delta.dmstorad())) /
                    (math.cos(location.latitude.dmstorad()) *
                    math.cos(self.delta.dmstorad())))*180/math.pi
        tsg0 = AngleHMS(AstroDateTime(time.date+(0, 0, 0)).get_gmst()).hmstodeg()
        return AngleDeg((self.alpha.hmstodeg() +
                                hour_angle +
                                location.longitude.dmstodeg() -
                                tsg0) /
                                1.002737909).degtohms()



"""This module contains scripts functions
"""
import math
import numpy as np
import pkg_resources

from matplotlib import pyplot as plt
from matplotlib.offsetbox import OffsetImage, AnnotationBbox, TextArea

from astro_toolbox.angle.hms import AngleHMS
from astro_toolbox.angle.dms import AngleDMS
from astro_toolbox.time.core import AstroDateTime
from astro_toolbox.coordinates.location import Location
from astro_toolbox.coordinates.equatorial import Equatorial
from astro_toolbox.query.ephemeris import Horizons
from astro_toolbox.query.weather import OpenMeteo


PATH = pkg_resources.resource_filename('astro_toolbox', 'query/weather_icons/')

def offset_image(name, axis):
    """Add icons from query/weather_icons on plot.

    Parameters
    ----------
    name : str
        icon name.
    axis : Object
        Pyplot axis object.

    Returns
    -------
    Object
        Pyplot image object.
    """
    path = PATH + f'{name}.png'
    img = plt.imread(path)
    image = OffsetImage(img, zoom=0.5)
    image.image.axes = axis
    return image

def wmotoimage(code, axis):
    """Function to convert WMO code to icon.

    Parameters
    ----------
    code : str
        WMO code as str.
    axis : Object
        Axis object.

    Returns
    -------
    Object
        offset_image Object.
    """
    if code == '00':
        icon = offset_image('sunny', axis)
    elif code == '01':
        icon = offset_image('mainly_clear', axis)
    elif code == '02':
        icon = offset_image('partly_cloudy', axis)
    elif code == '03':
        icon = offset_image('cloudy', axis)
    elif code[0] == '4':
        icon = offset_image('rain_mix', axis)
    elif code[0] == '5':
        icon = offset_image('drizzle', axis)
    elif code[0] == '6' or code == 23:
        icon = offset_image('rain', axis)
    elif code[0] == '7' or code == '21':
        icon = offset_image('snow', axis)
    elif code[0] == ('8'):
        icon = offset_image('showers', axis)
    elif code[0] == ('9'):
        icon = offset_image('thunderstorm', axis)
    else:
        icon = offset_image('na', axis)
    return icon

def winddirectiontoimage(wind_direction, axis):
    """Funcion to convert wind direction to icon.

    Parameters
    ----------
    wind_direction : float
        Wind direction
    axis : Object
        Axis object.

    Returns
    -------
    Object
        offset_image Object.
    """
    if 0 <= wind_direction <= 22.5 or 337.5 <= wind_direction <= 360:
        icon = offset_image('direction_up', axis)
    elif 22.5 < wind_direction < 67.5:
        icon = offset_image('direction_up_left', axis)
    elif 67.5 < wind_direction < 112.5:
        icon = offset_image('direction_left', axis)
    elif 112.5 < wind_direction < 157.5:
        icon = offset_image('direction_down_left', axis)
    elif 157.5 < wind_direction < 202.5:
        icon = offset_image('direction_down', axis)
    elif 202.5 < wind_direction < 247.5:
        icon = offset_image('direction_down_right', axis)
    elif 247.5 < wind_direction < 292.5:
        icon = offset_image('direction_right', axis)
    elif 292.5 < wind_direction < 337.5:
        icon = offset_image('direction_up_right', axis)
    return icon

def plot_weather(date, location, bounds, axis):
    """Plot weather on airmass map

    Parameters
    ----------
    date : str | tuple
        Plotting date.
    location : Location
        Location object.
    bounds : tuple
        Tuple containing observation bounds.
    axis : Object
        Pyplot axis object.
    """
    hours = list(np.mod(np.arange(bounds[0],bounds[1]), 24))
    weather_forecast = OpenMeteo(location=location)
    annotation_box = AnnotationBbox(TextArea('Weather\nforecasts',
                                            textprops=dict(rotation = 90, ha = 'center')),
                                            (0, 0),
                                            xybox=(-85, -55),
                                            frameon=False,
                                            xycoords='data',
                                            boxcoords="offset points",
                                            pad=0)
    axis.add_artist(annotation_box)
    weather_units = [TextArea('Temperature °C',
                            textprops=dict(size=6)),
                    TextArea('Humidity %',
                            textprops=dict(size=6)),
                    TextArea('Precipitation mm',
                            textprops=dict(size=6)),
                    TextArea('Wind Speed km/h',
                            textprops=dict(size=6)),
                    TextArea('Wind Direction',
                            textprops=dict(size=6))]
    for j, item in enumerate(weather_units):
        annotation_box = AnnotationBbox(item,
                                        (0, 0),
                                        xybox=(-40, -35 - 10 * j),
                                        frameon=False,
                                        xycoords='data',
                                        boxcoords="offset points",
                                        pad=0)
        axis.add_artist(annotation_box)
    last_time = 0
    for time in hours:
        if time - last_time < 0:
            date = AstroDateTime(date).get_gregorian(1)
        last_time = time
        datetime = date + (time, 0, 0)
        weather = [offset_image('na', axis)]
        weather += [TextArea(f'{weather_forecast.get_temperature(datetime)}',
                            textprops=dict(size=6)),
                    TextArea(f'{weather_forecast.get_humidity(datetime)}',
                            textprops=dict(size=6)),
                    TextArea(f'{weather_forecast.get_precipitation(datetime)}',
                            textprops=dict(size=6)),
                    TextArea(f'{weather_forecast.get_wind_speed(datetime)}',
                            textprops=dict(size=6))]
        weather.append(winddirectiontoimage(
                        weather_forecast.get_wind_direction(datetime), axis))
        weather[0] = wmotoimage(weather_forecast.get_wmo(datetime), axis)
        for j, item in enumerate(weather):
            annotation_box = AnnotationBbox(item,
                                            (hours.index(time) * 10, 0),
                                            xybox=(0., -25 - 10 * j),
                                            frameon=False,
                                            xycoords='data',
                                            boxcoords="offset points",
                                            pad=0)
            axis.add_artist(annotation_box)

def sun_impact(lines, site, date, bounds):
    """Function which plots sun impact.

    Parameters
    ----------
    lines : int
        number of lines plot.
    site : Location
        Location object.
    date : tuple | str
        Date to plot sun impact.
    bounds : list
        List of length 2 containing lower and upper time bounds.

    Returns
    -------
    list
        List of matplotlib.pyplot Rectangles objects.
    """
    ut_time = AstroDateTime(date)
    sun = Horizons('Sun', date, site)
    sun = Equatorial(name='Sun',
                    alpha=sun.get_equatorial_coord()[0],
                    delta=sun.get_equatorial_coord()[1],
                    magnitude=sun.get_magnitude)
    sun_rectangles_list = []
    for i, time in enumerate(np.arange(bounds[0],bounds[1],0.1)):
        time = (int(time),(time-int(time))*60,0)
        ut_time = AstroDateTime(ut_time.date+time)
        altitude = (AngleDMS(sun.to_horizontal(ut_time.get_lst(site), site)[1]).dmstodeg())
        if altitude <= 0:
            opacity = 1 + altitude/18
            if altitude < -18:
                opacity = 0
        elif altitude > 0:
            opacity = 1
        sun_rectangles_list.append(plt.Rectangle((i, 0), 1, lines,
                                                alpha=opacity, color='w', linewidth=3))
    return sun_rectangles_list

def moon_times(lines, site, date, bounds):
    """Moon observational impact.

    Parameters
    ----------
    lines : int
        number of lines plot.
    site : Location
        Location object.
    date : tuple | str
        Date to plot moon impact.
    bounds : list
        List of length 2 containing lower and upper time bounds.
    """
    moon = Horizons('Moon', date, site)
    moon = Equatorial(name='Moon',
                    alpha=moon.get_equatorial_coord()[0],
                    delta=moon.get_equatorial_coord()[1],
                    magnitude=0)
    rise_time = moon.calculate_rise_time(site,
                                        date,
                                        altitude_0=0)
    set_time = moon.calculate_set_time(site,
                                    date,
                                    altitude_0=0)
    rise_time = rise_time[0] + rise_time[1]/60 + rise_time[2]/3600
    set_time = set_time[0] + set_time[1]/60 + set_time[2]/3600
    lower_bound = [0] * (int((bounds[1] - bounds[0])/24)+2)
    upper_bound = [0] * (int((bounds[1] - bounds[0])/24)+2)
    for i in range(0,int((bounds[1] - bounds[0])/24)+2):
        lower_bound[i] += 24 * i
        upper_bound[i] += 24 + 24 * i
    lower_bound[0] = bounds[0]
    upper_bound[-1] = bounds[1]
    for lower, upper in zip(lower_bound, upper_bound):
        if lower < set_time + 24 * int(lower/24) < upper:
            plt.plot([(set_time + 24 * int(lower/24)-bounds[0])*10]*2,
                    [0, lines], '--', color='grey', linewidth=2, label='Moonset')
        if lower < rise_time + 24 * int(lower/24) < upper:
            plt.plot([(rise_time + 24 * int(lower/24)-bounds[0])*10]*2,
                    [0, lines], ':', color='grey', linewidth=2, label='Moonrise')
    handles, labels = plt.gca().get_legend_handles_labels()
    by_label = dict(zip(labels, handles))
    plt.legend(by_label.values(),
                by_label.keys(),
                bbox_to_anchor=(1.05, 1),
                loc=2, borderaxespad=0.,
                facecolor='k',
                labelcolor='red')

def airmas_map(object_dict,
                site: Location,
                date: tuple|str,
                bounds: tuple=(18,31)):
    """Airmass calculations.

    Parameters
    ----------
    object_dict : dict
        Dictionary containing object name ans theirs associated equatorial class.
    site : Location
        Location object.
    date : tuple | str
        Date to plot airmass map.
    bounds : tuple
        List of length 2 containing lower and upper time bounds.
    """
    ut_time = AstroDateTime(date)
    bounds = list(bounds)
    if bounds[1] <= bounds[0]:
        bounds[1] = bounds[1] + 24
    airmasses = np.empty((len(object_dict), np.shape(np.arange(bounds[0],bounds[1],0.1))[0]))
    alpha_visible = np.ones((np.shape(airmasses)[0], np.shape(airmasses)[1]))
    for j, key in enumerate(object_dict):
        object_dict[key].compute_on_date_coord(year = ut_time.get_year()+
                                                (ut_time.get_month()-1.0)/12)
        for i, time in enumerate(np.arange(bounds[0],bounds[1],0.1)):
            time = (int(time),(time-int(time))*60,0)
            ut_time = AstroDateTime(ut_time.date+time)
            airmasses[j][i] = object_dict[key].calculate_airmass(
                                ut_time.get_lst(site), site)
            if airmasses[j][i] >= 40:
                alpha_visible[len(object_dict)-j-1][i] = 0
    fig, axis = plt.subplots(figsize=(8.3, 11.7))
    moon_times(len(object_dict), site, date, bounds)
    mesh = axis.pcolormesh(airmasses[::-1],
                    cmap='jet',
                    shading='flat',
                    alpha=alpha_visible,
                    vmin = 1,
                    vmax = 3.5)
    for rect in sun_impact(len(object_dict), site, date, bounds):
        axis.add_patch(rect)
    fig.colorbar(mesh, ax=axis, shrink=0.5)
    plt.xticks(np.arange(0,len(np.arange(bounds[0],bounds[1],0.1)),10),
                                np.mod(np.arange(bounds[0],bounds[1]), 24))
    plt.yticks(np.arange(0.5,len(object_dict),1),
                [item[1].name + ' v='
                + str(item[1].magnitude)
                for item in reversed(object_dict.items())])
    plot_weather(ut_time.date, site, bounds, axis)
    axis.tick_params(left = False)
    axis.grid(axis = 'x', color = 'k')
    axis.set_title(f'Airmass: {ut_time.get_day():02.0f}/{ut_time.get_month():02.0f}'
            +f'/{ut_time.get_year():04.0f} UT @ {site.name}')
    axis.set_facecolor('k')
    fig.tight_layout()
    return fig

def polarstar_plt_northern(location: Location, datetime: tuple | str=None):
    """Polaris Polar Finder position for northern hemisphere.

    Parameters
    ----------
    site : Location
        Location object.
    datetime : tuple | str
        Date and time to plot Polaris position.
    """
    ut_time = AstroDateTime(datetime)
    fig, axis = plt.subplots(figsize=(5, 5), num='Polaris position')
    axis.add_patch(plt.Circle((0,0), 36, color='r', fill=False))
    axis.add_patch(plt.Circle((0,0), 40, color='r', linestyle='--', fill=False))
    axis.add_patch(plt.Circle((0,0), 44, color='r', fill=False))
    axis.plot([-21, 21], [0, 0], color='r')
    axis.plot([0, 0], [-21, 21], color='r')
    axis.text(-1, -50, '0', color='r')
    axis.text(47, -2, '6', color='r')
    axis.text(-3, 47, '12', color='r')
    axis.text(-52, -2, '18', color='r')
    polar_star = Equatorial(name='polaris',
                    alpha=(2, 32, 08.50),
                    delta=(+89, 16, 11.6),
                    magnitude=1.95)
    polar_star.compute_on_date_coord(year = ut_time.get_year()+
                                        (ut_time.get_month()-1.0)/12)
    polar_star_hour_angle = AngleHMS(polar_star.get_hourangle(
                                    ut_time.get_lst(location)))
    polar_star_distance = (90 - polar_star.delta.dmstodeg())*60
    x_polar_star, y_polar_star = (polar_star_distance*math.sin(polar_star_hour_angle.hmstorad()),
                                -polar_star_distance*math.cos(polar_star_hour_angle.hmstorad()))
    axis.plot(x_polar_star,y_polar_star, 'ow')
    axis.set_aspect('equal', adjustable='box')
    axis.axis('off')
    axis.set_xlim([-60, 60])
    axis.set_ylim([-60, 60])
    axis.text(20, 60, f"HA = {polar_star_hour_angle}", color='r')
    axis.text(20, 55, f"d (') = {round(polar_star_distance, 2)}", color='r')
    fig.tight_layout()
    fig.patch.set_facecolor('k')
    plt.show()

def polarstar_plt_southern(location: Location, datetime: tuple=None):
    """Polaris Polar Finder position for southern hemisphere

    Parameters
    ----------
    site : Location
        Location object.
    datetime : tuple | str
        Date and time to plot Polaris australis position.
    """
    ut_time = AstroDateTime(datetime)
    fig, axis = plt.subplots(figsize=(8, 8), num='Polaris position')
    axis.add_patch(plt.Circle((0,0), 60, color='r', fill=False))
    axis.add_patch(plt.Circle((0,0), 65, color='r', linestyle='--', fill=False))
    axis.add_patch(plt.Circle((0,0), 70, color='r', fill=False))
    axis.plot([-21, 21], [0, 0], color='r')
    axis.plot([0, 0], [-21, 21], color='r')
    axis.text(-1, -50, '0', color='r')
    axis.text(47, -2, '6', color='r')
    axis.text(-3, 47, '12', color='r')
    axis.text(-52, -2, '18', color='r')
    octans_constellation =[[
    Equatorial(name='\u03C4 Oct',
                alpha=(23, 27, 43.30),
                delta=(-87, 29, 11.0),
                magnitude=5.50),
    Equatorial(name='CG Oct',
                alpha=(22, 44, 41.05),
                delta=(-88, 49, 19.3),
                magnitude=6.50),
    Equatorial(name='\u03C3 Oct',
                alpha=(21, 7, 37.48),
                delta=(-88, 57, 27.9),
                magnitude=5.45),
    Equatorial(name='\u03C7 Oct',
                alpha=(18, 54, 16.53),
                delta=(-87, 36, 19.0),
                magnitude=5.25)], [], []]
    for star in octans_constellation[0]:
        star.compute_on_date_coord(year = ut_time.get_year()+
                                    (ut_time.get_month()-1.0)/12)
        octans_constellation[1].append((star.delta.dmstodeg() + 90)*60)
    for star in octans_constellation[0]:
        octans_constellation[2].append(
            AngleHMS(star.get_hourangle(ut_time.get_lst(location))).hmstorad())
    x_octans_stars = []
    y_octans_stars = []
    for i in range(len(octans_constellation[0])):
        x_octans_stars.append(-octans_constellation[1][i]*math.sin(octans_constellation[2][i]))
        y_octans_stars.append(octans_constellation[1][i]*math.cos(octans_constellation[2][i]))
    axis.plot(x_octans_stars, y_octans_stars, 'r', marker='o')
    axis.plot(x_octans_stars[2], y_octans_stars[2], 'ow')
    axis.text(x_octans_stars[0]-10, y_octans_stars[0]-10, '\u03C4', color='r')
    axis.set_aspect('equal', adjustable='box')
    axis.axis('off')
    axis.set_xlim([-160, 160])
    axis.set_ylim([-160, 160])
    fig.tight_layout()
    fig.patch.set_facecolor('k')
    plt.show()



"""
This module init all the public classes.
"""
from astro_toolbox.time.core import AstroDateTime
from astro_toolbox.angle.degrees import AngleDeg
from astro_toolbox.angle.radians import AngleRad
from astro_toolbox.angle.dms import AngleDMS
from astro_toolbox.angle.hms import AngleHMS
from astro_toolbox.coordinates.equatorial import Equatorial
from astro_toolbox.coordinates.horizontal import Horizontal
from astro_toolbox.coordinates.location import Location
from astro_toolbox.coordinates.solar_system import Ephemeris
from astro_toolbox.query.ephemeris import Horizons
from astro_toolbox.query.catalogs import Simbad
from astro_toolbox.query.weather import OpenMeteo

from astro_toolbox.query.ephemeris import DICT_OBJECTS

__all__ = [
    "AstroDateTime",
    "AngleDeg",
    "AngleRad",
    "AngleDMS",
    "AngleHMS",
    "Equatorial",
    "Horizontal",
    "Location",
    "Ephemeris",
    "Simbad",
    "Horizons",
    "OpenMeteo",
]

__version__ = "0.1.0"



"""
This module contains AngleHMS class.
"""
import math
from astro_toolbox.utils.strparser import angle_parser
class AngleHMS():
    """AngleHMS define a HMS angle with its conversions.

    Attributes
    ----------
    anglevalue : tuple
        The angle value as floats in tuple.
    """
    def __init__(self, anglevalue: tuple | str):
        """Constructor method

        Parameters
        ----------
        anglevalue : tuple | str
            The angle value in hms as a tuple of float or a string (``ddhddmdds`` or ``dd:dd:dd``).
        """
        if isinstance(anglevalue, str):
            anglevalue = angle_parser(anglevalue)
        self.anglevalue = anglevalue

    def __repr__(self):
        """Representative method.

        Returns
        -------
        string
            Return a class representative string.
        """
        return f'{self.anglevalue[0]:02.0f}h{self.anglevalue[1]:02.0f}m{self.anglevalue[2]:05.2f}s'

    def hmstodeg(self):
        """HMS to Degrees converting method.
        This method return angle in degrees from angle in HMS.

        Returns
        -------
        float
            The angle value in degrees
        """
        return ((self.anglevalue[0] +
            self.anglevalue[1]/60 +
            self.anglevalue[2]/3600)*180/12)

    def hmstorad(self):
        """HMS to Radians converting method.
        This method return angle in radians from angle in HMS.

        Returns
        -------
        float
            The angle value in radians.
        """
        return self.hmstodeg() * math.pi/180



"""This module contains Horizons class.
"""
import re
import json
import urllib.request as urllib

from astro_toolbox.time.core import AstroDateTime
from astro_toolbox.coordinates.location import Location

DICT_OBJECTS = {
                'Sun': 10,
                'Mercury': 199,
                'Venus': 299,
                'Moon': 301,
                'Mars': 499,
                'Jupiter': 599,
                'Saturn': 699,
                'Uranus': 799,
                'Neptune': 899,
                'Pluto': 999
                }

class Horizons():
    """JPL Horizons ephemeris request and parsing.

    Attributes
    ----------
    name : str | int
        The object name or integer reference according to JPL Horizons.
    datetime : AstroDateTime
        Date and time as AstroDateTime class.
    location : Location
        Observer location as Location class.
    object_data : list
        List which contains results from Horizons.
    """
    def __init__(self, object_name: str | int, datetime: tuple | str, location: Location):
        """Constructor method

        Parameters
        ----------
        object_name : str | int
            The object name or integer reference according to JPL Horizons.
        datetime : tuple | str
            Date and time as tuple or str (``dd:dd:dd.dd`` or ``ddhddmdd.dds``).
        location : Location
            Observer location as Location class.
        """
        self.name = object_name
        self.datetime = AstroDateTime(datetime)
        self.location = location
        self.object_data = self._get_data()

    def _get_data(self):
        """JPL data query method.

        Returns
        -------
        list
            A list containing the request result.

        Raises
        ------
        ValueError
            Unknown object.
        """
        if isinstance(self.name, int):
            object_ref = self.name
        if self.name.lower() in [key.lower() for key in DICT_OBJECTS]:
            keys = list(key for key in DICT_OBJECTS)
            idx = list(key.lower() for key in DICT_OBJECTS).index(self.name.lower())
            object_ref = DICT_OBJECTS[keys[idx]]
        else:
            raise ValueError ("Object doesn't exist verify value")
        dict_parameters = {'COMMAND': object_ref,
                            'SITE_COORD': 500,
                            'TLIST': f"{self.datetime.get_jd()}"
                            }
        link = ("https://ssd.jpl.nasa.gov/api/horizons.api?" +
                "format=json&OBJ_DATA=%27NO%27&QUANTITIES=%271,9%27")
        for key, value in dict_parameters.items():
            link = link + f"&{key}=%27{value}%27"
        request=urllib.Request(link)
        with urllib.urlopen(request) as response:
            result = json.loads(response.read().decode('utf-8'))['result']
        return re.split(r"\*+",result)

    def get_equatorial_coord(self):
        """Get equatorial coordinates from JPL Horizons.

        Returns
        -------
        tuple
            Equatorial coordinates right_ascension as HMS tuple and declination as DMS tuple.
        """
        result = self.object_data[6]
        result = re.split(r"\s",re.split(r"\s{2,}",result)[2])
        return ((int(result[0]), int(result[1]), float(result[2])),
                (int(result[3]), int(result[4]), float(result[5])))

    def get_magnitude(self):
        """Get magnitude from JPL Horizons.

        Returns
        -------
        float
            Magnitude.
        """
        result = self.object_data[6]
        result = re.split(r"\s",re.split(r"\s{2,}",result)[3])[0]
        return float(result)

    def get_name(self):
        """Get object name from JPL Horizons.

        Returns
        -------
        str
            Object name.
        """
        result = self.object_data[2]
        result = re.findall(r"Target body name: (\w+)", result)
        return result[0]



"""This module contains AstroDateTime class.
"""
import math
import datetime
from astro_toolbox.coordinates.location import Location
from astro_toolbox.utils.strparser import angle_parser
class AstroDateTime():
    """This module contains AngleRad class

    Attributes
    ----------
    ut_time : tuple | str
        Date and time as tuple or string in format (2000/01/01/12:00:00)
    """
    def __init__(self, ut_time: tuple | str = None):
        """Constructor method

        Parameters
        ----------
        ut_time : tuple
            Tuple of the date and time in format (year,month,day,hour,minute,second)
        """
        if ut_time is None:
            ut_time = datetime.datetime.now(datetime.timezone.utc).timetuple()
        if ut_time is not None and isinstance(ut_time, str):
            ut_time = angle_parser(ut_time)
        if ut_time is not None and len(ut_time) == 3:
            ut_time += (12, 0, 0)
        self.date = (ut_time[0], ut_time[1], ut_time[2])
        self.time = (ut_time[3]%24, ut_time[4], ut_time[5])
        if ut_time is not None and ut_time[3] > 23:
            ut_time = self.get_gregorian(0) + ut_time[3:6]
        self.date = (int(ut_time[0]), int(ut_time[1]), int(ut_time[2]))
        self.time = (int(ut_time[3])%24, int(ut_time[4]),int(ut_time[5]))

    def __repr__(self):
        """Representative method.

        Returns
        -------
        string
            Return a class representative string.
        """
        return (f'{self.date[0]:02d}-{self.date[1]:02d}-{self.date[2]:02d}'+
                f'T{self.time[0]:02d}:{self.time[1]:02d}:{self.time[2]:02d}')

    def get_year(self):
        """Get year

        Returns
        -------
        int
            Year.
        """
        return self.date[0]

    def get_month(self):
        """Get month

        Returns
        -------
        int
            Month.
        """
        return self.date[1]

    def get_day(self):
        """Get day.

        Returns
        -------
        int
            Day.
        """
        return self.date[2]

    def get_time(self):
        """Get time.

        Returns
        -------
        tuple
            Time in format (hour,minute,second).
        """
        return self.time

    def get_jd(self, delta: int = 0):
        """Get julian day with possibility of delta days from USNO formula.

        .. math:: JD=367year-\\frac{7(year+\\frac{month+9}{12})}{4}>+\\frac{275month}{9}+
                day+1721013.5+\\frac{UT}{24}-\n
        .. math:: 0.5sign(100year+month-190002.5)+0.5

        Parameters
        ----------
        delta : int
            Delta in days.

        Returns
        -------
        float
            Julian day.
        """
        julian_day = (367 * self.date[0] -
        int((7 * (self.date[0] +
        int((self.date[1] + 9) / 12.0))) / 4.0) +
        int((275 * self.date[1]) / 9.0) +
        self.date[2] + 1721013.5 +
        (self.time[0] + self.time[1] / 60.0 + self.time[2] / 3600) / 24.0 -
        0.5 * math.copysign(1, 100 * self.date[0] + self.date[1] - 190002.5) + 0.5)
        return julian_day + delta

    def get_gregorian(self, delta: int = 0):
        """Get gregorian date with possibility of delta days from USNO formulas.

        .. math:: g_1 = julian_day + 68569
        .. math:: g_2 = <\\frac{4g_1}{146097}>
        .. math:: g_1 = g_1 - <\\frac{146087g_2 + 3}{4}>
        .. math:: g_3 = <\\frac{4000(g_1 + 1)}{1461001}>
        .. math:: g_1 = g_1 - <\\frac{1461g_3}{4}> + 31
        .. math:: g_4 = <\\frac{80g_3}{2447}>
        .. math:: day = g_1 - <\\frac{2447g_4}{80}>
        .. math:: g_1 = <\\frac{g_4}{11}
        .. math:: month = g_4 + 2 - 12g_1
        .. math:: year = 100(g_2 - 49) + g_3 + g_1

        Parameters
        ----------
        delta : int, optional
            Delta days, by default 0.

        Returns
        -------
        tuple
            Tuple of int containing year month and month day.
        """
        julian_day = int(self.get_jd(delta) + 0.5)
        g_1 = julian_day + 68569
        g_2 = int(4 * g_1 / 146097)
        g_1 = g_1 - int((146097 * g_2 + 3) / 4)
        g_3 = int(4000 * (g_1 + 1) / 1461001)
        g_1 = g_1 - int(1461 * g_3 / 4) + 31
        g_4 = int(80 * g_1 / 2447)
        day = g_1 - int(2447 * g_4 / 80)
        g_1 = int(g_4 / 11)
        month = g_4 + 2 - 12 * g_1
        year = 100 * (g_2 - 49) + g_3 + g_1
        return year, month, day

    def get_year_day(self):
        """Get day of year.

        .. math:: N1 = floor(275 * month / 9)
        .. math:: N2 = floor((month + 9) / 12)
        .. math:: N3 = (1 + floor((year - 4 * floor(year / 4) + 2) / 3))
        .. math:: N = N1 - (N2 * N3) + day - 30

        Returns
        -------
        int
            Day of year.
        """
        n_1 = math.floor(275 * self.get_month() / 9)
        n_2 = math.floor((self.get_month() + 9) / 12)
        n_3 = 1 + math.floor((self.get_year() - 4 * math.floor(self.get_year() / 4) + 2) / 3)
        return n_1 - (n_2 * n_3) + self.get_day() - 30

    def get_gmst(self):
        """Get Greenwich mean sidereal time with USNO formula.

        .. math:: gmst=mod(18.697375+24.065709824279(JD-2451545), 24)

        Returns
        -------
        tuple
            Greenwich mean sidereal time.
        """
        julian_day = self.get_jd()
        gmst = (18.697375 + 24.065709824279 * (julian_day - 2451545)) % 24
        gmst_mm = (gmst - int(gmst))*60
        gmst_ss = (gmst_mm - int(gmst_mm))*60
        gmst_hh = int(gmst)
        gmst_mm = int(gmst_mm)
        return (gmst_hh,gmst_mm,gmst_ss)

    def get_lst(self, location: Location):
        """Get local mean sidereal time with USNO formula.

        .. math:: lst=gmst+\\frac{\\lambda}{15}

        Parameters
        ----------
        location : Location
            Observer location.

        Returns
        -------
        tuple
            Local mean sidereal time.
        """
        gmst = self.get_gmst()
        longitude = location.longitude.dmstodeg()/15
        lst = (gmst[0]+gmst[1]/60+gmst[2]/3600) + longitude
        if lst < 0:
            lst = lst + 24
        lst_mm = (lst - int(lst))*60
        lst_ss = (lst_mm - int(lst_mm))*60
        lst_hh = int(lst)
        lst_mm = int(lst_mm)
        return (lst_hh,lst_mm,lst_ss)



"""This module contains scripts functions.
"""
import pathlib
import pkg_resources
from rich.progress import track

from astro_toolbox.coordinates.equatorial import Equatorial
from astro_toolbox.query.ephemeris import Horizons
from astro_toolbox.query.catalogs import Simbad

from astro_toolbox.query.ephemeris import DICT_OBJECTS

PATH = pkg_resources.resource_filename('astro_toolbox', '/coordinates/data/')

def read_observatory_program(input_file: pathlib.Path):
    """Read an observatory program from a list.

    Parameters
    ----------
    input_file : pathlib.Path
        Input file path.

    Returns
    -------
    object_list : list
        Objects list.
    """
    with open(input_file, 'r', encoding="utf-8") as file:
        object_list = file.readlines()
        object_list = [n.replace('\n', '') for n in object_list]
    if '' in object_list:
        object_list.remove('')
    object_list.sort(key=str.casefold)
    return object_list

def get_multiple_informations(object_list, site, datetime, bounds):
    """Getting multiple objects equatorial information.
    """
    bounds = list(bounds)
    if bounds[1] <= bounds[0]:
        bounds[1] = bounds[1] + 24
    object_dict = {}
    for name in track(object_list, description="Querying information..."):
        try:
            if name.lower() in list(key.lower() for key in DICT_OBJECTS):
                obj = Horizons(name, datetime, site)
            else:
                obj = Simbad(name)
            object_dict[name] = Equatorial(alpha=obj.get_equatorial_coord()[0],
                                        delta=obj.get_equatorial_coord()[1], name=name,
                                        magnitude=obj.get_magnitude())
        except ValueError:
            pass
    return object_dict



"""This module creates the console interface.
"""
import re
import pathlib
import json
import logging
import pkg_resources
import click
from matplotlib.backends.backend_pdf import PdfPages
from rich.progress import track

from astro_toolbox.angle.degrees import AngleDeg
from astro_toolbox.angle.hms import AngleHMS
from astro_toolbox.time.core import AstroDateTime
from astro_toolbox.coordinates.location import Location
from astro_toolbox.coordinates.equatorial import Equatorial
from astro_toolbox.query.ephemeris import Horizons
from astro_toolbox.query.catalogs import Simbad
from astro_toolbox.query.weather import OpenMeteo
from astro_toolbox.scripts.planning import read_observatory_program
from astro_toolbox.scripts.planning import get_multiple_informations
from astro_toolbox.scripts.plots import airmas_map
from astro_toolbox.scripts.plots import polarstar_plt_northern
from astro_toolbox.scripts.plots import polarstar_plt_southern

from astro_toolbox.query.ephemeris import DICT_OBJECTS

PATH = pkg_resources.resource_filename('astro_toolbox', 'coordinates/data/')
PATH_2 = pkg_resources.resource_filename('astro_toolbox', 'query/weather_icons/')

@click.group()
@click.option(
    "-v",
    "--verbose",
    count=True,
    default=0,
    help="-v for DEBUG",
)
def cli(verbose):
    """Display verbose for command line

    Parameters
    ----------
    verbose : int
        Display verbose or not with logging levels
    """
    logging_levels = {1: logging.CRITICAL,
                          2: logging.ERROR,
                          3: logging.INFO,
                          4: logging.DEBUG}
    if verbose:
        logging.basicConfig(level = logging_levels[verbose],
                            format='%(levelname)s:astro_toolbox: %(message)s')

@cli.command("airmass")
@click.option("-d", "--date",
            type=click.STRING,
            default=None,
            help='-d, --date the date default is None if None, today date')
@click.option("-l", "--location",
            type=click.STRING,
            default=None,
            help='-l --location the site name default is None if None last site used')
@click.option("--bounds",
            nargs=2,
            type=click.INT,
            default=((18, 7)),
            help='--bounds to set airmass calculation hours bounds, default=(18, 31)')
@click.option("-o", "--output",
            type=click.STRING,
            default='',
            help='-o --output to set output path, default=\'\'')
@click.argument('input_file_objects',
            type=click.STRING,
            nargs=-1)
def airmass_map_command(input_file_objects, output, location, date, bounds):
    """Airmass calculations.

    Parameters
    ----------
    input_file_objects : str
        Multiple objects, directory or file (must contain '/' for path).
    output : str
        Outpt path, must contain '/' at the end.
    location : str
        Saved site name.
    date : str
        Specified date.
    bounds : tuple
        Tuple which contains night beginning bound and night ending bound.
    """
    inputpath = False
    if len(input_file_objects) == 0:
        inputpath = pathlib.Path('observations.lst')
    else:
        if '/' in input_file_objects[0]:
            inputpath = pathlib.Path(input_file_objects[0])
            if inputpath.is_dir():
                inputpath = pathlib.Path(input_file_objects[0]+'observations.lst')
        else:
            object_list = list(input_file_objects)
            object_list.sort(key=str.casefold)
    if inputpath is not False:
        object_list = read_observatory_program(inputpath)
    ut_time = AstroDateTime(date)
    site = Location(location)
    object_dict = get_multiple_informations(object_list, site, date, bounds)
    temporary_dict = {}
    pdf = PdfPages(pathlib.Path(output +
                    f'Airmass_Map_{ut_time.get_day():02.0f}_{ut_time.get_month():02.0f}'
                    +f'_{ut_time.get_year():04.0f}_@_{site.name}.pdf'))
    for count, key in track(enumerate(object_dict), description='Plotting maps...'):
        temporary_dict.update({key:object_dict[key]})
        if count == len(object_dict)-1 or ((count+1)%50 == 0):
            pdf.savefig(airmas_map(temporary_dict, site, date, bounds))
            temporary_dict = {}
    del temporary_dict
    pdf.close()
    logging.info(f'Airmass: {ut_time.get_day():02.0f}/{ut_time.get_month():02.0f}'
                +f'/{ut_time.get_year():04.0f} UT @ {site.name}')

@cli.command('info')
@click.argument('objects_list',
                nargs=-1,
                type=click.STRING)
@click.option("-d", "--datetime",
            type=click.STRING,
            default=None,
            help='-t, --time the date default is None if None, now')
@click.option("-l", "--location",
            type=click.STRING,
            default=None,
            help='-l --location the site name default is None if None last site used')
def info_command(objects_list, datetime, location):
    """Getting celestial objects information.

    Parameters
    ----------
    objects_list : str
        Multiple objects list.
    time : str
        Date and time
    location : str
       Saved site name.
    """
    for object_name in objects_list:
        site = Location(name=location)
        if object_name.lower() in [key.lower() for key in DICT_OBJECTS]:
            obj = Horizons(object_name, datetime, site)
            alpha, delta = obj.get_equatorial_coord()
            obj_name = obj.get_name()
            obj_magnitude = obj.get_magnitude()
        else:
            obj = Simbad(object_name)
            alpha, delta = obj.get_equatorial_coord()
            obj_name = obj.get_name()
            obj_magnitude = obj.get_magnitude()
        coord = Equatorial(alpha=alpha, delta=delta, name=obj_name, magnitude=obj_magnitude)
        coord.compute_on_date_coord(AstroDateTime(datetime).get_year())
        gamma = AstroDateTime(datetime).get_lst(site)
        click.echo(f'{coord}' +
                f' HA = {AngleHMS(coord.get_hourangle(gamma=gamma))}'
                f' X = {coord.calculate_airmass(gamma=gamma, location=site):.2f}' +
                f' @ {site.name} {AstroDateTime(datetime)}')

@cli.command('location')
@click.argument('location_name')
@click.option("-a", "--add",
            count=True,
            default=0,
            help='-a, --add to add location')
@click.option('-d', '--delete',
            count=True,
            default=0,
            help='-d, --delete to delete location')
@click.option('-u', '--update',
            count=True,
            default=0,
            help='-u, --update to update location')
def location_command(location_name, add, delete, update):
    """Command to make action son saved locations

    Parameters
    ----------
    location_name : str
        Location name
    add : bool
        Command to add site
    delete : bool
        Command to delete site
    update : bool
        Command to update site
    """
    def _verify_coords(latitude, longitude, elevation):
        """Function to verify coords

        Parameters
        ----------
        latitude : str
            Latitude string
        longitude : str
            Longitude string
        elevation : float
            Elevation float

        Returns
        -------
        list
            Coords list

        Raises
        ------
        ValueError
            Latitude or Longitude is not in necessary format
        ValueError
            Elevation is not a number
        """
        coords = [latitude, longitude]
        for count, value in enumerate(coords):
            value_flag = True
            if ('°' and "'") in value:
                coord_str = re.split(r"[°']",value)[:3]
                coords[count] = tuple(float(coord_val) for coord_val in coord_str)
                value_flag = False
            elif '.' in value and not ('°' and "'") in value:
                coords[count] = AngleDeg(float(value)).degtodms
                value_flag = False
            elif value_flag is False and (value != 'None'):
                raise ValueError(f"{value} Latitude or longitude is not in format 0.0 or 0°0'0''")
            if not isinstance(elevation, float):
                raise ValueError((f"{value} Elevation value is not a number"))
        return coords
    if add:
        latitude = str(input("Enter latitude in degrees (00.00) or in dms (00°00'00\"): "))
        if '.' in latitude and ('°' and "'") not in latitude:
            latitude = AngleDeg(float(latitude)).degtodms()
        longitude = str(input("Enter longitude in degrees (00.00) or in dms (00°00'00\"): "))
        if '.' in longitude and ('°' and "'") not in longitude:
            longitude = AngleDeg(float(longitude)).degtodms()
        elevation = float(input("Enter elevation in meters: ") or 0)
        coords = (latitude, longitude, elevation)
        Location(location_name, coords[0], coords[1], elevation).save_site()

    if delete:
        Location(name=location_name).delete_site()

    if update:
        click.echo('Enter only values to update')
        latitude = (str(input("Enter latitude in degrees (00.00) or in dms (00°00'00''): ")
                    or None))
        longitude = (str(input("Enter longitude in degrees (00.00) or in dms (00°00'00''): ")
                    or None))
        elevation = (float(input("Enter elevation in meters: ")
                    or 'nan'))
        coords = _verify_coords(latitude=latitude, longitude=longitude, elevation=elevation)
        Location(location_name, coords[0], coords[1], elevation).update_site()

    if location_name == 'list':
        with open(PATH  + 'sites.json', encoding="utf-8") as json_file:
            dict_sites = json.load(json_file)
        name_list = list(dict_sites.keys())
        count = 1
        for name in name_list:
            click.echo(f"{count}: " + f"{name}: latitude: {dict_sites[name]['latitude']} "+
                f"longitude: {dict_sites[name]['longitude']} "+
                f"elevation = {dict_sites[name]['elevation']} m")
            count = count + 1

@cli.command('polaris')
@click.option("-t", "--time",
            type=click.STRING,
            default=None,
            help='-t, --time the date and time default is None if None, now')
@click.option("-l", "--location",
            type=click.STRING,
            default=None,
            help='-l --location the site name default is None if None last site used')
def polaris_command(location, time):
    """Polaris position calculation.

    Parameters
    ----------
    location : str
        Saved site name.
    time : str
        Date and time.
    """
    location = Location(location)
    if location.latitude.dmstodeg() > 0:
        polarstar_plt_northern(location, time)
    elif location.latitude.dmstodeg() < 0:
        polarstar_plt_southern(location, time)

@cli.command('weather')
@click.option("-l", "--location",
            type=click.STRING,
            default=None,
            help='-l --location the site name default is None if None last site used')
@click.option(
    "-d",
    "--days",
    count=True,
    default=1)
def weather_command(location, days):
    """Weather forecasts displayed from Open-Meteo

    Parameters
    ----------
    location : str
        Saved site name.
    days : int
        Number of days displayed.
    """
    last_time = '0000-00-00T00:00:00'
    print(f'weather_forecasts @ {Location(location)}')
    weather_forecasts = OpenMeteo(Location(location))
    for time in weather_forecasts.data['hourly']['time'][:24*days]:
        time += ':00'
        if int(time[8:10]) != int(last_time[8:10]):
            print('=' * 174)
            print(f"|{'Time (UT)':^25}|{'Temperature (°C)':^20}|" +
                f"{'Humidity (%)':^20}|{'Precipitation (mm)':^20}|" +
                f"{'Wind Speed (km/h)':^20}|{'Wind Direction (°)':^20}|" +
                f"{'Cloud Coverage (%)':^20}|{'WMO':^20}|")
            print('=' * 174)
        else:
            print('-' * 174)
        print(f'|{time:^25}|{weather_forecasts.get_temperature(time):^20}|' +
              f'{weather_forecasts.get_humidity(time):^20}|' +
              f'{weather_forecasts.get_precipitation(time):^20}|' +
              f'{weather_forecasts.get_wind_speed(time):^20}|' +
              f'{weather_forecasts.get_wind_direction(time):^20}|' +
              f'{weather_forecasts.get_cloud(time):^20}|' +
              f'{weather_forecasts.get_wmo(time):^20}|')
        last_time = time
    print('-' * 174)

if __name__ == '__main__':
    cli(False)



"""
This module contains AngleDeg class.
"""
import math
class AngleDeg():
    """AngleDeg define degrees angle with its conversions.

    Attributes
    ----------
    anglevalue: float
        The angle value in degrees.
    """
    def __init__(self, anglevalue: float):
        """Constructor method

        Parameters
        ----------
        anglevalue : float
            The angle value in degrees.
        """
        self.anglevalue = anglevalue

    def __repr__(self):
        """Representative method.

        Returns
        -------
        string
            Return a class representative string.
        """
        return f'{self.anglevalue}°'

    def degtorad(self):
        """Degrees to radians converting method.
        This method returns angle in radians from angle in degrees.

        Returns
        -------
        float
            The angle value in radians.
        """
        return self.anglevalue*math.pi/180

    def degtodms(self):
        """Degrees to DMS converting method.
        This method returns angle in DMS from angle in degrees.

        Returns
        -------
        tuple
            The angle values in DMS.
        """
        deg_dms_result = (math.copysign(int(self.anglevalue), self.anglevalue),)
        deg_dms_value = abs(self.anglevalue - int(self.anglevalue)) * 60
        deg_dms_result = deg_dms_result + (int(deg_dms_value),)
        deg_dms_value = abs(deg_dms_value - int(deg_dms_value)) * 60
        deg_dms_result = deg_dms_result + (deg_dms_value,)
        return deg_dms_result

    def degtohms(self):
        """Degrees to HMS converting method.
        This method returns angle in HMS from angle in degrees.

        Returns
        -------
        tuple
            The angle values in HMS.
        """
        deg_hms_value = (self.anglevalue*12/180)%24
        deg_hms_result = (int(deg_hms_value),)
        deg_hms_value = (deg_hms_value - int(deg_hms_value)) * 60
        deg_hms_result = deg_hms_result + (int(deg_hms_value),)
        deg_hms_value = (deg_hms_value - int(deg_hms_value)) * 60
        deg_hms_result = deg_hms_result + (deg_hms_value,)
        return deg_hms_result



"""This module contains Open_Meteo class.
"""
import json
import urllib.request as urllib
import math

from astro_toolbox.time.core import AstroDateTime
from astro_toolbox.coordinates.location import Location

class OpenMeteo():
    """OpenMeteo service request and parsing.

    Attributes
    ----------
    location : Location
            Observer location as Location class.
    model : str, optional
        Weather model (https://open-meteo.com/en/docs), by default 'best_match'
    """
    def __init__(self, location : Location, model : str = 'best_match'):
        """Constructor method

        Parameters
        ----------
        location : Location
            Observer location as Location class.
        model : str, optional
            Weather model (https://open-meteo.com/en/docs), by default 'best_match'
        """
        self.location = location
        self.pressure_level = self.get_pressure_level()
        self.model = model
        self.data = self._get_data()

    def __repr__(self):
        """Representative method.

        Returns
        -------
        str
            representative string
        """
        return (f'Temperature: {self.get_temperature()} °C' +
                f'Humidity: {self.get_humidity()} %' +
                f'Precipitation: {self.get_precipitation()} mm' +
                f'Wind Speed: {self.get_wind_speed()} km/h' +
                f'Wind Direction: {self.get_wind_direction()} °' +
                f'Cloud Coverage: {self.get_cloud()} %')

    def _get_data(self):
        """Open Meteo query method.

        Returns
        -------
        dict
            Dict containing all the datas requested.
        """
        link = 'https://api.open-meteo.com/v1/forecast?'
        link += (f'latitude={str(self.location.latitude.dmstodeg())}&' +
                f'longitude={str(self.location.longitude.dmstodeg())}&' +
                'hourly=' +
                f'temperature_{self.pressure_level}hPa,' +
                f'relativehumidity_{self.pressure_level}hPa,' +
                'precipitation_probability,' +
                'precipitation,' +
                'weathercode,' +
                'pressure_msl,' +
                f'cloudcover_{self.pressure_level}hPa,' +
                f'windspeed_{self.pressure_level}hPa,' +
                f'winddirection_{self.pressure_level}hPa')

        link += f'&models={self.model}&current_weather=false&past_days=1&forecast_days=16'
        request=urllib.Request(link)
        with urllib.urlopen(request) as response:
            result = json.loads(response.read().decode('utf-8'))
        result.pop('generationtime_ms')
        return result

    def _get_index(self, datetime: str | tuple):
        """Time searching method.

        Parameters
        ----------
        datetime : str | tuple
            Date and Time as tuple (YYYY,MM,DD,hh,mm,ss) or YYYY-MM-DDThh:mm:ss.

        Returns
        -------
        int
            Time index.
        """
        datetime = AstroDateTime(datetime)
        datetime_str = str(datetime)[:-5]+'00'
        if datetime_str not in self.data['hourly']['time']:
            return float('nan')
        return self.data['hourly']['time'].index(datetime_str)

    def get_pressure_level(self):
        """Pressure level computing method.

        .. math:: P = 1013.25(1 - \\frac{h}{44307.694})^{5.25530}

        Returns
        -------
        int
            10^2 rounded pressure level in hPa.
        """
        return int(round((1013.25 *
                (1 - self.location.elevation /
                44307.694) ** 5.25530), -2))

    def get_temperature(self, datetime: str | tuple = None):
        """Get temperature method.

        Parameters
        ----------
        datetime : str | tuple, optional
            Date and Time as tuple (``YYYY,MM,DD,hh,mm,ss`` or ``YYYY-MM-DDThh:mm:ss``),
            by default None.

        Returns
        -------
        float
            Temperature
        """
        index = self._get_index(datetime)
        if math.isnan(index):
            return float('nan')
        return self.data['hourly'][f'temperature_{self.pressure_level}hPa'][index]

    def get_humidity(self, datetime: str | tuple = None):
        """Get humidity method.

        Parameters
        ----------
        datetime : str | tuple, optional
            Date and Time as tuple (``YYYY,MM,DD,hh,mm,ss`` or ``YYYY-MM-DDThh:mm:ss``),
            by default None.

        Returns
        -------
        float
            Humidity
        """
        index = self._get_index(datetime)
        if math.isnan(index):
            return float('nan')
        return self.data['hourly'][f'relativehumidity_{self.pressure_level}hPa'][index]

    def get_dewpoint(self, datetime: str | tuple = None):
        """Get dewpoint method.

        Parameters
        ----------
        datetime : str | tuple, optional
            Date and Time as tuple (``YYYY,MM,DD,hh,mm,ss`` or ``YYYY-MM-DDThh:mm:ss``),
            by default None.

        Returns
        -------
        float
            Dewpoint
        """
        index = self._get_index(datetime)
        if math.isnan(index):
            return float('nan')
        return self.data['hourly']['dewpoint_2m'][index]

    def get_precipitation(self, datetime: str | tuple = None):
        """Get precipitation method.

        Parameters
        ----------
        datetime : str | tuple, optional
            Date and Time as tuple (``YYYY,MM,DD,hh,mm,ss`` or ``YYYY-MM-DDThh:mm:ss``),
            by default None.

        Returns
        -------
        float
            Precipitation
        """
        index = self._get_index(datetime)
        if math.isnan(index):
            return float('nan')
        return self.data['hourly']['precipitation'][index]

    def get_precipitation_probability(self, datetime: str | tuple = None):
        """Get precipitation probability method.

        Parameters
        ----------
        datetime : str | tuple, optional
            Date and Time as tuple (``YYYY,MM,DD,hh,mm,ss`` or ``YYYY-MM-DDThh:mm:ss``),
            by default None.

        Returns
        -------
        float
            Precipitation probability
        """
        index = self._get_index(datetime)
        if math.isnan(index):
            return float('nan')
        return self.data['hourly']['precipitation_probability'][index]

    def get_wmo(self, datetime: str | tuple = None):
        """Get World Meteorological Organization code method.

        Parameters
        ----------
        datetime : str | tuple, optional
            Date and Time as tuple (``YYYY,MM,DD,hh,mm,ss`` or ``YYYY-MM-DDThh:mm:ss``),
            by default None.

        Returns
        -------
        float
            WMO cod
        """
        index = self._get_index(datetime)
        if math.isnan(index):
            return float('nan')
        code = self.data['hourly']['weathercode'][index]
        return f'{code:02d}'

    def get_pressure(self, datetime: str | tuple = None):
        """Get pressure method.

        Parameters
        ----------
        datetime : str | tuple, optional
            Date and Time as tuple (``YYYY,MM,DD,hh,mm,ss`` or ``YYYY-MM-DDThh:mm:ss``),
            by default None.

        Returns
        -------
        float
            Pressure
        """
        index = self._get_index(datetime)
        if math.isnan(index):
            return float('nan')
        return self.data['hourly']['surface_pressure'][index]

    def get_msl_pressure(self, datetime: str | tuple = None):
        """Get sea level pressure method.

        Parameters
        ----------
        datetime : str | tuple, optional
            Date and Time as tuple (``YYYY,MM,DD,hh,mm,ss`` or ``YYYY-MM-DDThh:mm:ss``),
            by default None.

        Returns
        -------
        float
            Sea level pressure
        """
        index = self._get_index(datetime)
        if math.isnan(index):
            return float('nan')
        return self.data['hourly']['pressure_msl'][index]

    def get_cloud(self, datetime: str | tuple = None):
        """Get cloud coverage method.

        Parameters
        ----------
        datetime : str | tuple, optional
            Date and Time as tuple (``YYYY,MM,DD,hh,mm,ss`` or ``YYYY-MM-DDThh:mm:ss``),
            by default None.

        Returns
        -------
        float
            Cloud coverage
        """
        index = self._get_index(datetime)
        if math.isnan(index):
            return float('nan')
        return self.data['hourly'][f'cloudcover_{self.pressure_level}hPa'][index]

    def get_wind_speed(self, datetime: str | tuple = None):
        """Get wind speed method.

        Parameters
        ----------
        datetime : str | tuple, optional
            Date and Time as tuple (``YYYY,MM,DD,hh,mm,ss`` or ``YYYY-MM-DDThh:mm:ss``),
            by default None.

        Returns
        -------
        float
            Wind speed
        """
        index = self._get_index(datetime)
        if math.isnan(index):
            return float('nan')
        return self.data['hourly'][f'windspeed_{self.pressure_level}hPa'][index]

    def get_wind_direction(self, datetime: str | tuple = None):
        """Get wind direction method.

        Parameters
        ----------
        datetime : str | tuple, optional
            Date and Time as tuple (``YYYY,MM,DD,hh,mm,ss`` or ``YYYY-MM-DDThh:mm:ss``),
            by default None.

        Returns
        -------
        float
            Wind direction
        """
        index = self._get_index(datetime)
        if math.isnan(index):
            return float('nan')
        return self.data['hourly'][f'winddirection_{self.pressure_level}hPa'][index]
    def get_units(self):
        """Get units.

        Returns
        -------
        dict
            Dict containing units.
        """
        return self.data['hourly_units']



"""This module contains Ephemeris class.
"""
import math
import json
import pkg_resources

from astro_toolbox.angle.radians import AngleRad
from astro_toolbox.time.core import AstroDateTime

PATH = pkg_resources.resource_filename('astro_toolbox', 'coordinates/data/')

class Ephemeris():
    """Ephemeris class computing solar system objects positions in different referential.
    Orbital elements calculated by JPL.

    Attributes
    ----------
    name : str
        Planet name.
    n_centuries : float
        Number of centuries from J2000.
    orbital_elements : dict
        Dictionary which contains object orbital elements for current date.
    """
    def __init__(self, name: str, datetime: tuple | str):
        """Constructor method

        Parameters
        ----------
        name : str
            Planet name.
        datetime : tuple | str
            Date and time as tuple or str (``dd:dd:dd.dd`` or ``ddhddmdd.dds``).
        """
        self.name = name
        self.n_centuries = (AstroDateTime(datetime).get_jd() - 2451545)/36525
        if self.name.lower() == 'sun':
            self.orbital_elements = self._get_orbital_elements('EM Bary')
        else:
            self.orbital_elements = self._get_orbital_elements(self.name)

    def _get_orbital_elements(self, name: str):
        """Get on date orbital elements.

        Parameters
        ----------
        name : str
            Object name.

        Returns
        -------
        dict
            Dictionary which contains object orbital elements for current date.

        Raises
        ------
        ValueError
            Unknown Object.
        """
        with open(PATH  + 'orbital_elements.json', encoding="utf-8") as json_file:
            dict_orbital_elements = json.load(json_file)
        if name.lower() in [key.lower() for key in dict_orbital_elements.keys()]:
            keys = list(key for key in dict_orbital_elements)
            idx = list(key.lower() for key in dict_orbital_elements).index(name.lower())
            name = keys[idx]
            orbital_elements = dict_orbital_elements[name]
        else:
            raise ValueError ("Unknown object")
        for key, value in orbital_elements.items():
            if ',' in value:
                parsed_value = value.split(',')
                value = (float(parsed_value[0]) +
                        float(parsed_value[1]) * self.n_centuries)
                if key in ('L', 'I', 'longperi', 'longnode'):
                    value = value%360
            if key == 'a':
                value = value * 149597870.700
            orbital_elements[key] = float(value)
        orbital_elements.update({"perihelion": (orbital_elements['longperi'] -
                                    orbital_elements['longnode'])%360})
        orbital_elements.update({"M": (orbital_elements['L'] -
                                    orbital_elements['longperi'])%360})
        eccentric_anomaly = (orbital_elements['M'] -
                            (orbital_elements['e']*180/math.pi) *
                            math.sin(orbital_elements['M']*math.pi/180))
        delta_eccentric_anomaly = 1
        while abs(delta_eccentric_anomaly) > 1e-6:
            delta_mean_anomaly = (orbital_elements['M'] - (
                                eccentric_anomaly -
                                (orbital_elements['e']*180/math.pi) *
                                math.sin(eccentric_anomaly*math.pi/180)))
            delta_eccentric_anomaly = (delta_mean_anomaly / (
                                    1 - orbital_elements['e'] *
                                    math.cos(eccentric_anomaly*math.pi/180)))
            eccentric_anomaly = (eccentric_anomaly +
                                delta_eccentric_anomaly)
        orbital_elements.update({"E": eccentric_anomaly%360})
        return orbital_elements

    def compute_earth_position(self):
        """Sun position from earth.
        Distance and true anomaly are computed by _compute_true_anomaly_distance method.

        lonearth = v + w

        .. math:: xs = r * cos(lonsun)
        .. math:: ys = r * sin(lonsun)

        Returns
        -------
        tuple
            Tuple containing sun position from earth
        """
        earth_orbital_elements = self._get_orbital_elements('EM Bary')
        true_anomaly, radius = self.compute_true_anomaly_distance(**earth_orbital_elements)
        lonearth = (true_anomaly*180/math.pi + earth_orbital_elements['perihelion'])
        x_sun = -radius * math.cos(lonearth*math.pi/180)
        y_sun = -radius * math.sin(lonearth*math.pi/180)
        return x_sun, y_sun

    def compute_true_anomaly_distance(self, **orbital_elements):
        """Compute object distance and true anomaly.

        .. math:: xv = a(cos(E) - e)
        .. math:: yv = a\\sqrt{1 - e^2} * sin(E)
        .. math:: v = atan2(yv, xv)
        .. math:: r = \\sqrt{xv^2 + yv^2}

        Returns
        -------
        tuple
            Tuple which contains true anomaly and distance.
        """
        x_true_anomaly = orbital_elements['a'] * (math.cos(orbital_elements['E']*math.pi/180) -
                                                        orbital_elements['e'])
        y_true_anomaly = orbital_elements['a'] * (math.sqrt(1 - orbital_elements['e']**2) *
                                                        math.sin(orbital_elements['E']*math.pi/180))
        return (math.atan2(y_true_anomaly, x_true_anomaly),
                math.sqrt(x_true_anomaly**2 + y_true_anomaly**2))

    def compute_ecliptic_position(self, **orbital_elements):
        """Compute the heliocentric ecliptic object position.

        .. math:: xh = r*(cos(\\Omega)*cos(v+\\omega)-sin(\\Omega)*sin(v+\\omega)*cos(I))
        .. math:: yh = r*(sin(\\Omega)*cos(v+\\omega)+cos(\\Omega)*sin(v+\\omega)*cos(I))
        .. math:: zh = r*(sin(v+\\omega)*sin(I))

        Returns
        -------
        tuple
            Tuple which contains object position around the sun.
        """
        true_anomaly, radius = self.compute_true_anomaly_distance(**orbital_elements)
        true_longitude = (true_anomaly*180/math.pi + orbital_elements['perihelion'])
        x_ecliptic = radius * (math.cos(orbital_elements['longnode']*math.pi/180) *
                                    math.cos(true_longitude*math.pi/180) -
                                    math.sin(orbital_elements['longnode']*math.pi/180) *
                                    math.sin(true_longitude*math.pi/180) *
                                    math.cos(orbital_elements['I']*math.pi/180))
        y_ecliptic = radius * (math.sin(orbital_elements['longnode']*math.pi/180) *
                                    math.cos(true_longitude*math.pi/180) +
                                    math.cos(orbital_elements['longnode']*math.pi/180) *
                                    math.sin(true_longitude*math.pi/180) *
                                    math.cos(orbital_elements['I']*math.pi/180))
        z_ecliptic = radius * (math.sin(true_longitude*math.pi/180) *
                                    math.sin(orbital_elements['I']*math.pi/180))
        return x_ecliptic, y_ecliptic, z_ecliptic

    def compute_geocentric_position(self, **orbital_elements):
        """Compute geocentric object position.

        .. math:: xg = xh + xs
        .. math:: yg = yh + ys
        .. math:: zg = zh

        Where, xs and ys are the geocentric earth coordinates.

        Returns
        -------
        tuple
            Tuple which contains the geocentric position.
        """
        (x_heliocentric,
        y_heliocentric,
        z_heliocentric) = self.compute_ecliptic_position(**orbital_elements)
        x_sun, y_sun = self.compute_earth_position()
        x_geocentric = x_heliocentric + x_sun
        y_geocentric = y_heliocentric + y_sun
        z_geocentric = z_heliocentric
        return x_geocentric, y_geocentric, z_geocentric

    def compute_equatorial_position(self, **orbital_elements):
        """Compute equatorial object position.

        For the sun:

        .. math:: xe = xs
        .. math:: ye = ys * cos(ecl)
        .. math:: ze = ys * sin(ecl)

        For the moon:

        .. math:: xe = r*(cos(\\Omega)*cos(v+\\omega)-sin(\\Omega)*sin(v+\\omega)*cos(I))
        .. math:: ye = r*(sin(\\Omega)*cos(v+\\omega)+cos(\\Omega)*sin(v+\\omega)*cos(I))
        .. math:: ze = r*(sin(v+\\omega)*sin(I))

        For the planets:

        .. math:: xe = xg
        .. math:: ye = yg * cos(ecl) - zg * sin(ecl)
        .. math:: ze = yg * sin(ecl) + zg * cos(ecl)

        Returns
        -------
        tuple
            Tuple which contains the geocentric position.
        """
        ecliptic_obliquity = self.calculate_ecliptic_obliquity()*math.pi/180
        if self.name.lower() == 'sun':
            x_sun, y_sun = self.compute_earth_position()
            x_equatorial = x_sun
            y_equatorial = y_sun * math.cos(ecliptic_obliquity)
            z_equatorial = y_sun * math.sin(ecliptic_obliquity)
        else:
            if self.name.lower() == 'moon':
                (x_geocentric,
                y_geocentric,
                z_geocentric) = self.compute_ecliptic_position(**orbital_elements)
            else:
                (x_geocentric,
                y_geocentric,
                z_geocentric) = self.compute_geocentric_position(**orbital_elements)
        x_equatorial = x_geocentric
        y_equatorial = (math.cos(ecliptic_obliquity) * y_geocentric -
                        math.sin(ecliptic_obliquity) * z_geocentric)
        z_equatorial = (math.sin(ecliptic_obliquity) * y_geocentric +
                        math.cos(ecliptic_obliquity) * z_geocentric)
        return(x_equatorial, y_equatorial, z_equatorial)

    def calculate_ecliptic_obliquity(self):
        """Calculate on date ecliptic obliquity.

        Returns
        -------
        float
            Ecliptic obliquity in degrees.
        """
        return (23.439279444444445 -
                0.013010213611111111 * self.n_centuries**1 -
                5.0861111111111115e-08 * self.n_centuries**2 +
                5.565e-07 * self.n_centuries**3 -
                1.6e-10 * self.n_centuries**4 -
                1.2055555555555555e-11 * self.n_centuries**5)

    def get_equatorial_coord(self):
        """Calculate equatorial coordinates.

        .. math:: RA = atan2(ye, xe)
        .. math:: DEC = atan2(ze, \\sqrt{xe^2 + ye^2})

        Returns
        -------
        tuple
            Tuple which contains right-ascension as HMS tuple and declination as DMS tuple.
        """
        (x_equatorial,
        y_equatorial,
        z_equatorial) = self.compute_equatorial_position(**self.orbital_elements)
        right_ascencion = math.atan2(y_equatorial, x_equatorial)
        declination = math.atan2(z_equatorial, math.sqrt(x_equatorial**2 +
                                                         y_equatorial**2))
        return (AngleRad(right_ascencion).radtohms(),
                AngleRad(declination).radtodms())

    def get_magnitude(self):
        """Get solar system objects magnitude.

        Returns
        -------
        float
            Solar system object magnitude.
        """
        solar_system_objects_magnitude = {'sun': -26.83,
                                          'mercury': 0.23,
                                          'venus': -4.14,
                                          'moon': -12.90,
                                          'mars': 0.71,
                                          'jupiter': -2.20,
                                          'saturn': 0.46,
                                          'uranus': 6.03,
                                          'Neptune': 7.78}
        return solar_system_objects_magnitude[self.name.lower()]



"""This module contains Horizontal class.
"""
import math
from astro_toolbox.angle.hms import AngleHMS
from astro_toolbox.angle.dms import AngleDMS
from astro_toolbox.angle.radians import AngleRad
from astro_toolbox.angle.degrees import AngleDeg
from astro_toolbox.coordinates.location import Location

class Horizontal():
    """This class represent horizontal coordinate system.

    Attributes
    ----------
    azimuth : AngleDMS
        Object azimuth.
    altitude : AngleDMS
        Object altitude.
    name : str
        Object name.
    magnitude : float
        Object magnitude.
    """
    def __init__(self, azimuth:
                tuple | str,
                altitude: tuple | str,
                name: str = None,
                magnitude: float = None):
        """Constructor method

        Parameters
        ----------
        azimuth : tuple | str
           Object azimuth as tuple or string (``dd:dd:dd.dd`` or ``dd°dd'dd.dd"``).
        altitude : tuple | str
            Object altitude as tuple or string (``dd:dd:dd.dd`` or ``dd°dd'dd.dd"``).
        name : str, optional
            Object name, by default None.
        magnitude : float
            Object magnitude, by default is None.
        """
        self.name = name
        self.azimuth = AngleDMS(azimuth)
        self.altitude = AngleDMS(altitude)
        self.magnitude = magnitude

    def __repr__(self):
        """Representative method

        Returns
        -------
        string
            Return a class representative string
        """
        return f'{self.name}: A = {self.azimuth} h = {self.altitude} v = {self.magnitude}'

    def calculate_airmass(self):
        """Airmass calculation method
        The airmass is calculate with the Pickering (2002) formula from DIO,
        The International Journal of Scientific History vol. 12.

        .. math:: X = \\frac{1}{sin(h+\\frac{244}{165+47h^{1.1}})}

        Returns
        -------
        float
            Object airmass.
        """
        altitude = self.altitude
        if altitude < 0:
            return 40
        return abs(1/(math.sin(AngleDeg(altitude + 244/(165 + 47 * (altitude) ** 1.1)).degtorad())))

    def to_equatorial(self, gamma: tuple | str, location: Location):
        """Horizontal to equatorial converting method.

        .. math:: \\delta=sin^{-1}(sin \\Phi sin h-cos \\Phi cos h cos A)

        .. math:: \\alpha=sin^{-1}(\\frac{-cosh cosA}{cos\\delta})-\\gamma

        Parameters
        ----------
        gamma : AngleHMS
            Sidereal Time angle as tuple or string.
        location : Location
            observer location.

        Returns
        -------
        Equatorial
            Object equatorial coordinates.
        """
        gamma_angle =  AngleHMS(gamma)
        lat = location.latitude.dmstorad()
        delta = AngleRad(math.asin(math.sin(lat) *
                math.sin(self.altitude.dmstorad()) +
                math.cos(lat) *
                math.cos(self.altitude.dmstorad()) *
                math.cos(self.azimuth.dmstorad())))
        alpha = AngleRad(gamma_angle.hmstorad() - math.asin(-math.cos(self.altitude.dmstorad()) *
                math.sin(self.azimuth.dmstorad()) / math.cos(delta.anglevalue)))
        return alpha.radtohms(), delta.radtodms()



"""This module contains functionS to parse str.
"""
import re
def angle_parser(str_angle):
    """string angle parser.

    Parameters
    ----------
    str_anle : str
        The angle value as str in format (`dd:dd:dd.dd` `dd°dd'dd.dd"` `ddhddmdd.dds` or `dd.dd`)

    Returns
    -------
    tuple | float
        the angle in correct format depending on its unity
    """
    angle_list =  re.split(r"[\-:°'\"hmsT]",str_angle)
    if len(angle_list) == 1 and "." in angle_list[0]:
        return float(angle_list[0])
    angle = [int(val) for val in angle_list]
    if len(angle) == 3:
        angle[-1] = float(angle_list[-1])
    return tuple(angle)



"""This module contains Location class.
"""
import re
import json
import pkg_resources
from astro_toolbox.angle.dms import AngleDMS
from astro_toolbox.utils.strparser import angle_parser

PATH = pkg_resources.resource_filename('astro_toolbox', 'coordinates/data/')
class Location():
    """This class represents the observer location.

    Attributes
    ----------
    name : str
        Location name.
    latitude : tuple | str
        Location latitude.
    longitude : tuple | str
        Location longitude.
    elevation : float
        Location elevation.
    """
    def __init__(self, name: str = None, latitude: tuple | str = None,
                longitude: tuple | str = None, elevation: float = None):
        """Constructor method.

        Parameters
        ----------
        name : str
            Location name.
        latitude : tuple | str
            Location latitude tuple or str (``dd:dd:dd.dd`` or ``dd°dd'dd.dd"``).
        longitude : tuple | str
            Location longitude tuple or str (``dd:dd:dd.dd`` or ``dd°dd'dd.dd"``).
        elevation : float, optional
            Location elevation (m), by default 0.0.
        """
        if name is None:
            dict_site = self._get_site()
            name = list(dict_site.keys())[0]
        else:
            try:
                dict_site = self._get_site(name=name)
                name = list(dict_site.keys())[0]
            except ValueError:
                pass
        if latitude is None:
            latitude_str = re.split(r"[°'\"]",dict_site[name]['latitude'])[:3]
            latitude = (float(latitude_str[0]), float(latitude_str[1]), float(latitude_str[2]))
        if isinstance(latitude, str) and latitude is not None:
            latitude = angle_parser(latitude)
        if longitude is None:
            longitude_str = re.split(r"[°'\"]",dict_site[name]['longitude'])[:3]
            longitude = (float(longitude_str[0]), float(longitude_str[1]), float(longitude_str[2]))
        if isinstance(longitude, str) and longitude is not None:
            longitude = angle_parser(longitude)
        if elevation is None:
            elevation = dict_site[name]['elevation']
        self.name = name
        self.latitude = AngleDMS(latitude)
        self.longitude = AngleDMS(longitude)
        self.elevation = elevation
        self.current_site = {self.name:
                            {
                                'latitude': repr(self.latitude),
                                'longitude': repr(self.longitude),
                                'elevation': self.elevation
                            }
            }

    def __repr__(self):
        """Representative method

        Returns
        -------
        string
            Return a class representative string.
        """
        return (f'{self.name}: latitude: {self.latitude} '+
                f'longitude: {self.longitude} '+
                f'elevation = {self.elevation} m')

    def save_site(self):
        """Method to save current site in sites file.

        Raises
        ------
        KeyError
            The first level key already exist.
        """
        with open(PATH  + 'sites.json', encoding="utf-8") as json_file:
            dict_sites = json.load(json_file)
        if self.name.lower() not in [key.lower() for key in dict_sites]:
            new_dict_sites = self.current_site
            new_dict_sites.update(dict_sites)
            with open(PATH  + 'sites.json', 'w', encoding="utf-8") as json_file:
                json.dump(new_dict_sites, json_file, indent=4)
            return None
        raise ValueError("Site already exist, use `update_site` instead")

    def delete_site(self):
        """Method to delete current site from saved sites file.

        Raises
        ------
        KeyError
            The first level key doesn't exist.
        """
        with open(PATH  + 'sites.json', encoding="utf-8") as json_file:
            dict_sites = json.load(json_file)
        if self.name.lower() in list(key.lower() for key in dict_sites):
            keys = list(key for key in dict_sites)
            idx = list(key.lower() for key in dict_sites).index(self.name.lower())
            del dict_sites[keys[idx]]
            with open(PATH  + 'sites.json', 'w', encoding="utf-8") as json_file:
                json.dump(dict_sites, json_file, indent=4)
            return None
        raise ValueError("Site already exist, use `update_site` instead")
    def update_site(self):
        """Method to update current site in saved sites file.

        Raises
        ------
        KeyError
            The first level key doesn't exist.
        """
        with open(PATH  + 'sites.json', encoding="utf-8") as json_file:
            dict_sites = json.load(json_file)
        if self.name.lower() in list(key.lower() for key in dict_sites):
            keys = list(key for key in dict_sites)
            idx = [key.lower() for key in dict_sites].index(self.name.lower())
            for key in dict_sites[keys[idx]]:
                if self.current_site[self.name][key] is not None:
                    dict_sites[keys[idx]][key] = self.current_site[self.name][key]
            with open(PATH  + 'sites.json', 'w', encoding="utf-8") as json_file:
                json.dump(dict_sites, json_file, indent=4)
            return None
        raise ValueError("Site doesn't exist")

    def _get_site(self, name: str = None):
        """Method to get a site with it data.

        Parameters
        ----------
        name : str, optional
            The site name, by default None.

        Returns
        -------
        dict
            The location dictionary from its saved data in saved sites file.

        Raises
        ------
        KeyError
            The first level key doesn't exist.
        ValueError
            The site name is not given.
        """
        with open(PATH  + 'sites.json', encoding="utf-8") as json_file:
            dict_sites = json.load(json_file)
        if name is None:
            name = list(dict_sites.keys())[0]
        if name.lower() in [key.lower() for key in dict_sites.keys()]:
            keys = list(key for key in dict_sites)
            idx = list(key.lower() for key in dict_sites).index(name.lower())
            dict_site = dict_sites[keys[idx]]
            new_dict_sites = {keys[idx]: dict_sites.pop(keys[idx]), **dict_sites}
            with open(PATH  + 'sites.json', 'w', encoding="utf-8") as json_file:
                json.dump(new_dict_sites, json_file, indent=4)
            return {keys[idx]: dict_site}
        raise ValueError(f"{name} site doesn't exist")



"""Special functions.
"""
def range_w_last_val(start: float, end: float, step: float):
    """Adaptation of in-built python `range` function.

    Parameters
    ----------
    start : float
        Start float.
    end : float
        End float.
    step : float
        Step float.

    Yields
    ------
    float
        Range values.
    """
    idx = start
    while idx < end:
        yield idx
        idx = idx + step
    yield end



"""
This module contains AngleDMS class.
"""
import math
from astro_toolbox.utils.strparser import angle_parser
class AngleDMS():
    """AngleDMS define a dms angle with its conversions.

    Attributes
    ----------
    anglevalue : tuple
        The angle values as floats in tuple.
    """
    def __init__(self, anglevalue: tuple | str):
        """Constructor method

        Parameters
        ----------
        anglevalue : tuple | str
            The angle value in dms as a tuple of float or a string (`dd°dd'dd"` or `dd:dd:dd`).
        """
        if isinstance(anglevalue, str):
            anglevalue = angle_parser(anglevalue)
        self.anglevalue = anglevalue

    def __repr__(self):
        """Representative method.

        Returns
        -------
        str | None
            Return a class representative string.
        """
        return (f'{self.anglevalue[0]:+03.0f}°'+
                f'{self.anglevalue[1]:02.0f}\''+
                f'{self.anglevalue[2]:05.2f}\"')

    def dmstodeg(self):
        """DMS to degrees converting method.
        This method return angle in degrees from angle in DMS.

        Returns
        -------
        float
            The angle value in degrees.
        """
        return math.copysign(abs(self.anglevalue[0]) +
                self.anglevalue[1]/60 +
                self.anglevalue[2]/3600, self.anglevalue[0])

    def dmstorad(self):
        """DMS to Radians converting method.
        This method return angle in radians from angle in DMS.

        Returns
        -------
        float
            The angle value in radians.
        """
        return self.dmstodeg()*math.pi/180



"""
This module contains AngleRad class.
"""
import math
class AngleRad():
    """AngleRad define radians angle with its conversions.

    Attributes
    ----------
        anglevalue : float
            The angle value in radians.
    """
    def __init__(self, anglevalue: float):
        """Constructor method.

        Parameters
        ----------
        anglevalue : float
            The angle value in radians.
        """
        self.anglevalue = anglevalue

    def __repr__(self):
        """Representative method.

        Returns
        -------
        string
            Return a class representative string.
        """
        return f'{self.anglevalue}rad'

    def radtodeg(self):
        """Radians to Degrees converting method.
        This method returns angle in radians from angle in degrees.

        Returns
        -------
        float
            The angle value in degrees.
        """
        return self.anglevalue*180/math.pi

    def radtodms(self):
        """Radians to DMS converting method.
        This method returns angle in DMS from angle in radians.

        Returns
        -------
        tuple
            The angle values in DMS.
        """
        rad_dms_value = self.anglevalue*180/math.pi
        rad_dms_result = (math.copysign(int(rad_dms_value), rad_dms_value), )
        rad_dms_value = abs(rad_dms_value - int(rad_dms_value)) * 60
        rad_dms_result = rad_dms_result + (int(rad_dms_value),)
        rad_dms_value = abs(rad_dms_value - int(rad_dms_value)) * 60
        rad_dms_result = rad_dms_result + (rad_dms_value,)
        return rad_dms_result

    def radtohms(self):
        """Radians to HMS converting method.
        This method returns angle in HMS from angle in radians.

        Returns
        -------
        tuple
            The angle values in HMS.
        """
        rad_hms_value = (self.anglevalue*12/math.pi)%24
        rad_hms_result = (int(rad_hms_value),)
        rad_hms_value = (rad_hms_value - int(rad_hms_value)) * 60
        rad_hms_result = rad_hms_result + (int(rad_hms_value),)
        rad_hms_value = (rad_hms_value - int(rad_hms_value)) * 60
        rad_hms_result = rad_hms_result + (rad_hms_value,)
        return rad_hms_result



"""This module contains Simbad class.
"""
import urllib.request as urllib
import xmltodict

class Simbad():
    """This class allows to read astronomical catalogs over internet.

    Attributes
    ----------
    object_name : str

    """
    def __init__(self,object_name: str):
        """Constructor method.

        Parameters
        ----------
        object_name : str
            Astronomical object name.
        result : dict
            The object dictionary returned from Simbad.
        """
        object_name = object_name.replace(' ', '+')
        self.result = self._get_datas(object_name)

    def _get_datas(self, object_name: str):
        """Seraching datas method

        Parameters
        ----------
        object_name : str
            Astronomical object name.

        Returns
        -------
        dict
            The object dictionary returned from Simbad.
        """
        link = 'https://cds.unistra.fr/cgi-bin/nph-sesame/-oIfx?'+object_name
        request=urllib.Request(link)
        with urllib.urlopen(request) as response:
            result = xmltodict.parse(response.read().decode('utf-8'))['Sesame']['Target']
        if 'Resolver' not in result:
            raise ValueError("Object doesn't exist")
        return result

    def get_equatorial_coord(self):
        """Get RA/DEC object coordinates.

        Returns
        -------
        Tuple
            Tuple which contains two tuples RA/DEC coordinates.
        """
        alpha, delta = self.result['Resolver']['jpos'].split(' ')
        alpha = alpha.split(':')
        delta = delta.split(':')
        return((int(alpha[0]), int(alpha[1]), float(alpha[2])),
                (int(delta[0]), int(delta[1]), float(delta[2])))

    def get_name(self):
        """Get object name.

        Returns
        -------
        str
            Object name from Simbad.
        """
        return self.result['name']

    def get_magnitude(self):
        """Get visual magnitude.

        Returns
        -------
        float, None
            The object magnitude if available.
        """
        if 'mag' in self.result['Resolver']:
            flux_list = self.result['Resolver']['mag']
            if isinstance(flux_list, dict):
                flux_list = [flux_list]
            for flux in flux_list:
                if flux['@band'] == 'V':
                    return float(flux['v'])
        return None
